from revscoring import ScorerModel
from revscoring.datasources.revision_oriented import revision
from revscoring.features import Feature


def process_reversed_last_two_in_rev_id(rev_id):
    last_two = str(rev_id)[-2:]
    if len(last_two) == 1:
        return int(last_two + "0")
    else:
        return int("".join(reversed(last_two)))

reversed_last_two_in_rev_id = Feature(
    "revision.reversed_last_two_in_rev_id",
    process_reversed_last_two_in_rev_id,
    returns=int,
    depends_on=[revision.id]
)


class RevIdScorer(ScorerModel):
    """
    Implements a basic, testing scorer that predicts whether a revision ID's
    reversed last two digits are greater than 50.

    E.g. 974623 = 32 and 23754929 = 92
    """
    def __init__(self, version=None):
        super().__init__([reversed_last_two_in_rev_id], version=version)

    def score(self, feature_values):
        probability = feature_values[0] / 100

        if probability > 0.5:
            prediction = True
        else:
            prediction = False

        return {
            'prediction': prediction,
            'probability': {
                True: probability,
                False: 1 - probability
            }
        }

    def info(self):
        return {
            'type': "RevIDScorer",
            'version': self.version,
            'behavior': "Returns the last two digits in a rev_id as a score."
        }

    def format_info(self, format="str"):
        if format == 'str':
            return str(self.info())
        elif format == 'json':
            return self.info()
        else:
            raise TypeError("Format {0} not supported".format(format))

    @classmethod
    def from_config(cls, config, name, section_key='scorer_models'):
        section = config[section_key][name]
        return cls(**{k: v for k, v in section.items() if k != "class"})



import logging

from .score_cache import ScoreCache

logger = logging.getLogger(__name__)


class LRU(ScoreCache):

    def __init__(self, size=1024):
        try:
            import pylru
        except ImportError:
            raise ImportError("Could not find pylru.  This packages is " +
                              "required when using ores.score_caches.LRU.")

        self.lru = pylru.lrucache(size)

    def lookup(self, wiki, model, rev_id, version=None, cache=None):
        # Deterministic hash of cache values
        cache_hash = hash(tuple(sorted((cache or {}).items())))
        key = (wiki, model, rev_id, version, cache_hash)
        logger.debug("Looking up score at {0}".format(key))
        return self.lru[key]

    def store(self, wiki, model, rev_id, score, version=None, cache=None):
        # Deterministic hash of cache values
        cache_hash = hash(tuple(sorted((cache or {}).items())))
        key = (wiki, model, rev_id, version, cache_hash)
        logger.debug("Storing score at {0}".format(key))
        self.lru[key] = score

    @classmethod
    def from_config(cls, config, name, section_key="score_caches"):
        section = config[section_key][name]

        return cls(**{k: v for k, v in section.items() if k != "class"})



import os

from flask import render_template, send_file

from . import ui
from . import v1
from . import v2

PWD = os.path.dirname(os.path.abspath(__file__))
GEAR_FAVICON = os.path.join(PWD, "../static/favicon/gear/favicon.ico")


def configure(config, bp, score_processor):

    @bp.route("/", methods=["GET"])
    def index():
        return render_template("home.html")

    @bp.route("/favicon.ico", methods=["GET"])
    def favicon():
        # Tries to read from config or just loads the gear..
        return send_file(config['ores'].get('favicon', GEAR_FAVICON))

    bp = ui.configure(config, bp)
    bp = v1.configure(config, bp, score_processor)
    bp = v2.configure(config, bp, score_processor)

    return bp



from collections import namedtuple

from nose.tools import eq_

from ..logger import Logger


def test_logger():
    messages = []
    FakeLogger = namedtuple("Logger", ["debug"])
    logging_logger = FakeLogger(lambda m: messages.append(m))

    collector = Logger(logging_logger)
    collector.precache_request("foo", "bar", "0.0.1", 100)
    collector.scores_request("foo", "bar", "0.0.1", 50, 150)
    collector.datasources_extracted("foo", "bar", "0.0.1", 10, 25)
    collector.score_processed("foo", "bar", "0.0.1", 1.1)
    collector.score_cache_hit("foo", "bar", "0.0.1", 2)
    collector.score_errored("foo", "bar", "0.0.1")

    eq_(messages,
        ['precache_request: foo:bar:0.0.1 in 100 seconds',
         'scores_request: foo:bar:0.0.1 for 50 revisions in 150 seconds',
         'datasources_extracted: foo:bar:0.0.1 for 10 revisions in 25 seconds',
         'score_processed: foo:bar:0.0.1 in 1.1 seconds',
         'score_cache_hit: foo:bar:0.0.1',
         'score_cache_hit: foo:bar:0.0.1',
         'score_errored: foo:bar:0.0.1'])


def test_from_config():
    # Should throw a socket connection error and no others
    config = {
        'metrics_collectors': {
            'logger': {
                'class': 'ores.metrics_collectors.Logger'
            }
        }
    }
    Logger.from_config(config, 'logger')



from nose.tools import eq_, raises

from ..lru import LRU
from ..score_cache import ScoreCache


def test_lru():
    lru = LRU(2)

    cache_context = lru.context("foo", "bar")

    cache_context.store(1, "foo")
    cache_context.store(2, "bar")
    cache_context.store(3, "baz")  # 1 gets bumped

    eq_(cache_context.lookup(2), "bar")  # Moves 2 to the front
    cache_context.store(4, "fez")  # 3 gets bumped

    eq_(cache_context.lookup(2), "bar")  # Moves 2 to the front
    eq_(cache_context.lookup(4), "fez")  # Moves 4 to the front


@raises(KeyError)
def test_lru_cache():
    lru = LRU(2)

    cache_context = lru.context("foo", "bar")

    cache_context.store(1, "foo")
    cache_context.store(1, "cachedfoo", cache={"cachedvalue": 1})

    eq_(cache_context.lookup(1, cache={"cachedvalue": 1}), "cachedfoo")
    eq_(cache_context.lookup(1), "foo")

    # Raises KeyError
    cache_context.lookup(1, cache={"cachedvalue": 2})


def test_from_config():
    config = {
        'score_caches': {
            'mycache': {
                'class': 'ores.score_caches.LRU',
                'size': 128
            }
        }
    }

    cache = ScoreCache.from_config(config, "mycache")
    context = cache.context("foo", "bar")
    context.store(1, "foo")
    eq_(context.lookup(1), "foo")



import json
from collections import defaultdict

from flask.ext.jsonpify import jsonify


class CacheParsingError(Exception):
    pass


class ParamError(Exception):
    pass


def read_param(request, param, default=None, type=str):
    try:
        value = request.args.get(param, request.form.get(param))
        if value is None:
            return default
        else:
            return type(value)
    except (ValueError, TypeError) as e:
        raise ParamError("Could not interpret {0}. {1}"
                         .format(param, str(e)))


def read_bar_split_param(request, param, default=None, type=str):
    values = read_param(request, param, default=default)
    if values is None:
        return []

    try:
        return [type(value) for value in values.split("|")]
    except (ValueError, TypeError) as e:
        raise ParamError("Could not interpret {0}. {1}"
                         .format(param, str(e)))


def format_output(context, scores, model_info, warning=None, notice=None):
    """
    Formats a JSON blob of scores for API v2

    :Parameters:
        context: `str`
            Name of wiki
        scores : `dict`
            A JSONable dictionary of scores by revision ID and model.
        model_info : `dict`
            Information about mdoels
        warning: `dict`
            A warning if any
        notice: `dict`
            notice of deployment, etc. if any
    """
    output = defaultdict(dict)
    if notice:
        output['notice'] = notice
    if warning:
        output['warning'] = warning
    output['scores'] = {context: {}}
    for model in model_info:
        output['scores'][context][model] = model_info[model]
    output['scores'][context][model]['scores'] = {}
    for model in scores:
        output['scores'][context][model]['scores'] = {}
        for rev_id in scores[model]:
            output['scores'][context][model]['scores'][rev_id] = scores[model][rev_id]
    return jsonify(output)


def parse_injection(request, rev_id):
    """Parse values for features / datasources of interest."""
    cache = {}
    try:
        if 'inject' in request.values:
            cache = json.loads(request.values['inject'])

        for k, v in request.values.items():
            if k.startswith(("feature.", "datasource.")):
                cache[k] = json.loads(v)

        return {rev_id: cache} if len(cache) > 0 else None
    except Exception as e:
        raise CacheParsingError(e)



import logging

logger = logging.getLogger("ores.score_caches.score_cache")


class ScoreCache:

    def lookup(self, wiki, model, rev_id, version=None, cache=None):
        """
        Returns a pre-cached score
        """
        raise NotImplementedError()

    def store(self, wiki, model, rev_id, score, version=None, cache=None):
        """
        Caches a new score
        """
        raise NotImplementedError()

    def context(self, wiki, model, version=None):
        return Context(self, wiki, model, version=None)

    @classmethod
    def from_config(cls, config, name, section_key="score_caches"):
        try:
            import yamlconf
        except ImportError:
            raise ImportError("Could not find yamlconf.  This packages is " +
                              "required when using yaml config files.")
        logger.info("Loading ScoreCache '{0}' from config.".format(name))
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            Class = yamlconf.import_module(section['class'])
            return Class.from_config(config, name)


class Context:

    def __init__(self, scorer_cache, wiki, model, version=None):
        self.scorer_cache = scorer_cache
        self.wiki = str(wiki)
        self.model = str(model)
        self.version = version

    def lookup(self, rev_id, cache=None):
        return self.scorer_cache.lookup(self.wiki, self.model, rev_id,
                                        cache=cache, version=self.version)

    def store(self, rev_id, score, cache=None):
        return self.scorer_cache.store(self.wiki, self.model, rev_id, score,
                                       cache=cache, version=self.version)



import logging
import time
import traceback

from .. import errors
from ..metrics_collectors import Null
from ..score_caches import Empty
from ..util import jsonify_error

logger = logging.getLogger(__name__)


class ScoreProcessor(dict):

    def __init__(self, scoring_contexts, score_cache=None,
                       metrics_collector=None):
        super().__init__()
        self.update(scoring_contexts)
        self.score_cache = score_cache or Empty()
        self.metrics_collector = metrics_collector or Null()

    def score(self, context, model, rev_ids, caches=None,
              include_features=False, precache=False):
        version = self[context].version(model)
        start = time.time()
        if caches is not None:
            logger.debug("Scoring {0}:{1}:{2} with caches {3}"
                         .format(context, model, rev_ids, version, caches))

        try:
            scores = self._score(context, model, rev_ids, caches=caches,
                                 include_features=include_features)
        except errors.ScoreProcessorOverloaded:
            self.metrics_collector.score_processor_overloaded(
                context, model, version)
            raise

        duration = time.time() - start
        if not precache:
            self.metrics_collector.scores_request(context, model, version,
                                                  len(rev_ids), duration)
        else:
            self.metrics_collector.precache_request(context, model, version,
                                                    duration)

        return scores

    def _get_root_ds(self, context, model, rev_ids, caches=None):
        """
        Pure IO.  Batch extract root datasources for a set of features that the
        model needs.
        """
        if len(rev_ids) == 0:
            return {}
        rev_ids = set(rev_ids)
        scoring_context = self[context]
        version = scoring_context.version(model)

        start = time.time()
        roots = scoring_context.extract_roots(model, rev_ids, caches=caches)
        duration = time.time() - start
        logger.debug("Extracted root datasources for "
                     "{0}:{1}:{2}:{3} in {4} secs"
                     .format(context, model, version, rev_ids, duration))

        self.metrics_collector.datasources_extracted(context, model,
                                                     version, len(rev_ids),
                                                     duration)

        for error, cache in roots.values():
            if error is not None:
                logger.error("Errored while getting root datasource "
                             "for {0}:{1}:{2}:{3}"
                             .format(context, model, version, rev_ids))
                self.metrics_collector.score_errored(context, model, version)

        return roots

    def _process(self, context, model, cache, include_features=False):
        """
        Pure CPU.  Extract features from datasources in the cache and apply the
        model to arrive at a score.
        """
        scoring_context = self[context]
        version = scoring_context[model].version

        try:
            start = time.time()
            score, feature_vals = scoring_context.score(
                    model, cache, include_features=include_features)
            duration = time.time() - start
            logger.debug("Scoring {0}:{1}:{2} took {3} secs"
                         .format(context, model, version, duration))
            self.metrics_collector.score_processed(context, model, version,
                                                   duration)
        except:
            self.metrics_collector.score_errored(context, model, version)
            logger.error("Errored while scoring {0}:{1}:{2}"
                         .format(context, model, version))
            raise

        return score, feature_vals

    def _score_revision(self, context, model, rev_id, cache=None,
                        include_features=False):
        """
        Both IO and CPU.  Generates a single score or an error.
        """
        error, score_cache = self._get_root_ds(context, model, [rev_id],
                                               caches={rev_id: cache})[rev_id]

        if error is not None:
            raise error

        return self._process(context, model, score_cache,
                             include_features=include_features)

    def _store(self, context, model, rev_id, score, cache):
        version = self[context].version(model)

        self.score_cache.store(context, model, rev_id, score, cache=cache,
                               version=version)

    def _lookup_cached_scores(self, context, model, rev_ids, caches):
        version = self[context][model].version

        scores = {}
        for rev_id in rev_ids:
            cache = (caches or {}).get(rev_id, {})
            try:
                score = self.score_cache.lookup(context, model, rev_id,
                                                version=version, cache=cache)

                logger.debug("Found cached score for {0}:{1}:{2}{3}"
                             .format(context, model, rev_id,
                                     ":w/cache" if len(cache) > 0 else ""))

                self.metrics_collector.score_cache_hit(context, model, version)
                scores[rev_id] = score
            except KeyError:
                pass

        return scores

    @classmethod
    def from_config(cls, config, name, section_key="score_processors"):
        try:
            import yamlconf
        except ImportError:
            raise ImportError("Could not find yamlconf.  This packages is " +
                              "required when using yaml config files.")
        logger.info("Loading ScoreProcessor '{0}' from config.".format(name))
        section = config[section_key][name]
        if 'module' in section:
            return yamlconf.import_module(section['module'])
        elif 'class' in section:
            Class = yamlconf.import_module(section['class'])
            return Class.from_config(config, name)
        else:
            raise RuntimeError("No module or class to load.")


class SimpleScoreProcessor(ScoreProcessor):

    def _score(self, context, model, rev_ids, caches=None,
               include_features=False):
        rev_ids = set(rev_ids)

        # Look in the cache
        if not include_features:
            scores = self._lookup_cached_scores(context, model, rev_ids,
                                                caches)
            missing_ids = rev_ids - scores.keys()
        else:
            scores = {}
            missing_ids = rev_ids

        feature_maps = {}

        # Get the root datasources for the rest of the batch (IO)
        root_ds_caches = self._get_root_ds(context, model, missing_ids,
                                           caches=caches)

        # Extract features and generate scores (CPU)
        for rev_id, (error, cache) in root_ds_caches.items():
            if error is not None:
                scores[rev_id] = {'error': jsonify_error(error)}
            else:
                try:
                    score, feature_vals = \
                        self._process(context, model, cache,
                                      include_features=include_features)
                    scores[rev_id] = score
                    feature_maps[rev_id] = feature_vals
                    cache = (caches or {}).get(rev_id, {})
                    self._store(context, model, rev_id, score, cache)
                except Exception as error:
                    logger.debug(traceback.format_exc())
                    scores[rev_id] = {'error': jsonify_error(error)}

        return scores, feature_maps



import logging
import re
from hashlib import sha1
from urllib.parse import urlparse

import celery
import mwapi.errors
import revscoring.errors
from celery.signals import before_task_publish

from .. import errors
from ..metrics_collectors import MetricsCollector
from ..score_caches import ScoreCache
from ..util import jsonify_error
from .timeout import Timeout, TimeoutError

logger = logging.getLogger(__name__)

APPLICATIONS = []

DEFAULT_CELERY_QUEUE = "celery"


@before_task_publish.connect
def update_sent_state(sender=None, body=None, **kwargs):

    for application in APPLICATIONS:
        task = application.tasks.get(sender)
        backend = task.backend if task else application.backend

        logger.debug("Setting state to 'SENT' for {0}".format(body['id']))
        backend.store_result(body['id'], result=None, status="SENT")


class Celery(Timeout):

    def __init__(self, *args, application, queue_maxsize=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.application = application
        self.queue_maxsize = int(queue_maxsize) if queue_maxsize is not None \
                             else None

        self.redis = redis_from_url(self.application.conf.BROKER_URL)

        if self.queue_maxsize is not None and self.redis is None:
            logger.warning("No redis connection.  Can't check queue size")

        expected_errors = (revscoring.errors.RevisionNotFound,
                           revscoring.errors.DependencyError,
                           mwapi.errors.RequestError,
                           mwapi.errors.TimeoutError,
                           TimeoutError)

        @self.application.task(throws=expected_errors,
                               queue=DEFAULT_CELERY_QUEUE)
        def _process_task(context, model, cache, include_features):
            return Timeout._process(self, context, model, cache=cache,
                                    include_features=include_features)

        @self.application.task(throws=expected_errors,
                               queue=DEFAULT_CELERY_QUEUE)
        def _score_revision_task(context, model, rev_id, cache=None,
                                 include_features=False):
            return Timeout._score_revision(self, context, model, rev_id,
                                           cache=cache,
                                           include_features=include_features)

        APPLICATIONS.append(application)

        self._process_task = _process_task

        self._score_revision_task = _score_revision_task

    def _score_in_celery(self, context, model, rev_ids, caches,
                         include_features):
        scores = {}
        feature_maps = {}
        results = {}

        if len(rev_ids) == 0:
            return scores, feature_maps
        if len(rev_ids) == 1:  # Special case -- do everything in celery
            rev_id = rev_ids.pop()
            cache = (caches or {}).get(rev_id, {})
            id_string = self._generate_id(context, model, rev_id, cache)
            result = self._score_revision_task.apply_async(
                args=(context, model, rev_id),
                kwargs={'include_features': include_features, 'cache': cache},
                task_id=id_string
            )
            results[rev_id] = result
        else:  # Otherwise, try and batch
            # Get the root datasources for the rest of the batch (IO)
            root_ds_caches = self._get_root_ds(context, model, rev_ids,
                                               caches=caches)

            # Extract features and generate scores (CPU)
            for rev_id, (error, cache) in root_ds_caches.items():
                if error is not None:
                    scores[rev_id] = {'error': jsonify_error(error)}
                else:
                    injection_cache = (caches or {}).get(rev_id)
                    id_string = \
                        self._generate_id(context, model, rev_id,
                                          injection_cache)
                    result = self._process_task.apply_async(
                        args=(context, model, cache, include_features),
                        task_id=id_string
                    )
                    results[rev_id] = result

        # Process async results
        for rev_id, result in results.items():
            try:
                score, feature_vals = result.get(self.timeout)
                if feature_vals is not None:
                    feature_maps[rev_id] = feature_vals
                scores[rev_id] = score
                cache = (caches or {}).get(rev_id, {})
                self._store(context, model, rev_id, score, cache)
            except Exception as error:
                scores[rev_id] = {'error': jsonify_error(error)}

        return scores, feature_maps

    def _generate_id(self, context, model, rev_id, cache):
        scorer_model = self[context][model]
        version = scorer_model.version

        id_string = ":".join(str(v) for v in [context, model, rev_id, version])

        if cache is None or len(cache) == 0:
            return id_string
        else:
            cache_hash = self.hash_cache(cache)
            return id_string + ":" + cache_hash

    def _check_queue_full(self):
        # Check redis to see if the queue of waiting tasks is too big.
        # This is a hack to implement backpressure because celery doesn't
        # support it natively.
        # This will result in a race condition, but it should have OK
        # properties.
        if self.redis is not None:
            queue_size = self.redis.llen(DEFAULT_CELERY_QUEUE)
            if queue_size > self.queue_maxsize:
                message = "Queue size is too full {0}".format(queue_size)
                logger.warning(message)
                raise errors.ScoreProcessorOverloaded(message)

    def _score(self, context, model, rev_ids, caches=None,
               include_features=False):
        scores = {}
        feature_maps = {}
        self._check_queue_full()  # Raises ScoreProcessorOverloaded

        rev_ids = set(rev_ids)

        if not include_features:
            # Lookup scoring results that are currently in progress
            results = self._lookup_inprogress_results(context, model, rev_ids,
                                                      caches)
            missing_ids = rev_ids - results.keys()

            # Lookup scoring results that are in the cache
            scores = self._lookup_cached_scores(context, model, missing_ids,
                                                caches)
            missing_ids = missing_ids - scores.keys()
        else:
            results = {}
            missing_ids = rev_ids

        # Generate scores for missing rev_ids
        new_scores, new_feature_maps = \
            self._score_in_celery(context, model, missing_ids,
                                  caches=caches,
                                  include_features=include_features)
        scores.update(new_scores)
        feature_maps.update(new_feature_maps)

        # Gather results
        for rev_id in results:
            try:
                scores[rev_id], feature_maps[rev_id] = \
                    results[rev_id].get(self.timeout)
            except Exception as error:
                scores[rev_id] = {'error': jsonify_error(error)}

        # Return scores
        return scores, feature_maps

    def _lookup_inprogress_results(self, context, model, rev_ids, caches):
        scorer_model = self[context][model]
        version = scorer_model.version

        results = {}
        for rev_id in rev_ids:
            cache = (caches or {}).get(rev_id, {})
            id_string = self._generate_id(context, model, rev_id, cache)
            try:
                results[rev_id] = self._get_result(id_string)
                self.metrics_collector.score_cache_hit(context, model, version)
            except KeyError:
                pass

        return results

    def _get_result(self, id_string):

        # Try to get an async_result for an in_progress task
        result = self._score_revision_task.AsyncResult(task_id=id_string)
        logger.debug("Checking if {0} is already being processed [{1}]"
                     .format(repr(id_string), result.state))
        if result.state not in ("SENT", "STARTED", "SUCCESS"):
            raise KeyError(id_string)
        else:
            logger.debug("Found AsyncResult for {0}".format(repr(id_string)))
            return result

    @classmethod
    def hash_cache(cls, cache):
        sorted_tuple = tuple(sorted(cache.items()))
        return sha1(bytes(str(sorted_tuple), 'utf8')).hexdigest()

    @classmethod
    def from_config(cls, config, name, section_key="score_processors"):
        logger.info("Loading Celery '{0}' from config.".format(name))
        from ..scoring_contexts import ScoringContext

        section = config[section_key][name]

        scoring_contexts = {name: ScoringContext.from_config(config, name)
                            for name in section['scoring_contexts']}

        if 'score_cache' in section:
            score_cache = ScoreCache.from_config(config, section['score_cache'])
        else:
            score_cache = None

        if 'metrics_collector' in section:
            metrics_collector = \
                MetricsCollector.from_config(config,
                                             section['metrics_collector'])
        else:
            metrics_collector = None

        timeout = section.get('timeout')
        queue_maxsize = section.get('queue_maxsize')
        application = celery.Celery('ores.score_processors.celery')
        application.conf.update(**{k: v for k, v in section.items()
                                   if k not in ('class', 'timeout',
                                                'queue_maxsize')})
        application.conf.CELERY_CREATE_MISSING_QUEUES = True

        return cls(scoring_contexts, application=application, timeout=timeout,
                   score_cache=score_cache, queue_maxsize=queue_maxsize,
                   metrics_collector=metrics_collector)


PASS_HOST_PORT = re.compile(
    r"(?:\:(?P<password>[^@]+)@)?"
    r"(?P<host>[^:]+)?"
    r"(:(?P<port>[0-9]+))?"
)
"""
Matches <password>@<host>:<port>
"""


def redis_from_url(url):
    """
    Converts a redis URL used by celery into a `redis.Redis` object.
    """
    # Makes sure that we only try to import redis when we need
    # to use it
    import redis

    url = url or ""
    parsed_url = urlparse(url)
    if parsed_url.scheme != "redis":
        return None

    kwargs = {}
    match = PASS_HOST_PORT.match(parsed_url.netloc)
    if match.group('password') is not None:
        kwargs['password'] = match.group('password')
    if match.group('host') is not None:
        kwargs['host'] = match.group('host')
    if match.group('port') is not None:
        kwargs['port'] = int(match.group('port'))

    if len(parsed_url.path) > 1:
        # Removes "/" from the beginning
        kwargs['db'] = int(parsed_url.path[1:])

    return redis.StrictRedis(**kwargs)



"""
Runs a pre-caching server against an ORES instance by listening to an RCStream
and submitting requests for scores to the web interface for each revision as it
happens.

If ran by systemd, it uses watchdog to stay alive.

:Usage:
    precached -h | --help
    precached <stream-url> <ores-url> [--config=<path>] [--delay=<secs>]
                                      [--verbose]

:Options:
    -h --help        Prints this documentation
    <stream-url>     URL of an RCStream
    <ores-url>       URL of base ORES webserver
    --delay=<secs>   The delay between when an event is received and when the
                     request is sent to ORES. [default: 0]
    --config=<path>  The path to a yaml config file
                     [default: config]
    --verbose        Print debugging information
"""
import concurrent.futures
import glob
import logging
import os
import sys
import time
from collections import defaultdict

import docopt
import requests
import socketIO_client
import yamlconf

from .watchdog import notify_socket, watchdog_ping

logger = logging.getLogger(__name__)

AVAILABLE_EVENTS = {'edit', 'nonbot_edit'}


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    stream_url = args['<stream-url>']
    ores_url = args['<ores-url>']
    config_paths = os.path.join(args['--config'], "*.yaml")
    config = yamlconf.load(*(open(p) for p in
                             sorted(glob.glob(config_paths))))
    delay = float(args['--delay'])
    verbose = bool(args['--verbose'])
    notify = notify_socket()
    run(stream_url, ores_url, config, delay,
        notify, verbose)


def run(stream_url, ores_url, config, delay, notify, verbose):

    if verbose:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    logging.basicConfig(
        level=log_level,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )

    # Make requests and socketIO_client be quiet.  They are very noisy.
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("socketIO_client").setLevel(logging.ERROR)  # SHUT UP!
    requests.packages.urllib3.disable_warnings()

    if not notify:
        logger.info('Not being ran as a service, watchdog disabled')
    # Build a mapping of wikis and models from the configuration
    score_on = defaultdict(list)
    sp_name = config['ores']['score_processor']
    for context in config['score_processors'][sp_name]['scoring_contexts']:
        for model in config['scoring_contexts'][context].get('precache', []):
            precached_config = \
                config['scoring_contexts'][context]['precache'][model]

            events = precached_config['on']
            if len(set(events) - AVAILABLE_EVENTS) > 0:
                logger.error("{0} events are not available"
                             .format(set(events) - AVAILABLE_EVENTS))
                sys.exit(1)
            for event in precached_config['on']:
                score_on[(event, context)].append(model)
                logger.debug("Setting up precaching for {0} in {1} on {2}"
                             .format(model, context, event))

    def get_score(wiki, model, rev_id):
        url = ores_url + "/scores/" + wiki + "/" + model + \
              "/" + str(rev_id) + "/?precache=true"
        try:
            time.sleep(delay)
            start = time.time()
            requests.get(url, timeout=20, verify=True)
            logger.debug("GET {0} completed in {1} seconds."
                         .format(url, time.time() - start))
            if notify:
                watchdog_ping(*notify)
        except Exception as e:
            logger.error(str(type(e)) + ": " + str(e))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        class WikiNamespace(socketIO_client.BaseNamespace):
            def on_change(self, change):
                if change['type'] in ('new', 'edit'):
                    wikidb = change['wiki']
                    rev_id = change['revision']['new']
                    for model in score_on[('edit', wikidb)]:
                        executor.submit(get_score, wikidb, model, rev_id)

                    if not change.get('bot'):
                        for model in score_on[('nonbot_edit', wikidb)]:
                            executor.submit(get_score, wikidb, model, rev_id)

            def on_connect(self):
                logger.info("Connecting socketIO client to {0}."
                            .format(stream_url))

                # TODO: This should be limited to the wikis we actually care about
                # but we identify a wiki by it's dbname and this pattern matching
                # uses a domain name.
                self.emit('subscribe', '*')  # Subscribes to all wikis

        socketIO = socketIO_client.SocketIO(stream_url, 80)
        socketIO.define(WikiNamespace, '/rc')

        try:
            socketIO.wait(seconds=sys.maxsize)
        except KeyboardInterrupt:
            print("Keyboard interrupt detected.  Shutting down.")



from nose.tools import raises

from ..empty import Empty


def test_empty():

    empty = Empty()
    context = empty.context("foo", "bar", version="nope")
    context.store(1, "foo")


@raises(KeyError)
def test_key_error():

    empty = Empty()
    context = empty.context("foo", "bar", version="nope")
    context.store(1, "foo")

    context.lookup(1)



import logging
import time
import traceback

import stopit

from ..metrics_collectors import MetricsCollector
from ..score_caches import ScoreCache
from .score_processor import SimpleScoreProcessor

logger = logging.getLogger(__name__)


class Timeout(SimpleScoreProcessor):

    def __init__(self, *args, timeout=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.timeout = float(timeout) if timeout is not None else None

    def _process(self, context, model, cache, include_features=False):
        return timeout(super()._process, context, model, cache,
                       include_features=include_features, seconds=self.timeout)

    @classmethod
    def from_config(cls, config, name, section_key="score_processors"):
        logger.info("Loading Timeout '{0}' from config.".format(name))
        section = config[section_key][name]

        from ..scoring_contexts import ScoringContext

        section = config[section_key][name]

        scoring_contexts = {name: ScoringContext.from_config(config, name)
                            for name in section['scoring_contexts']}

        if 'score_cache' in section:
            score_cache = ScoreCache.from_config(config, section['score_cache'])
        else:
            score_cache = None

        if 'metrics_collector' in section:
            metrics_collector = \
                MetricsCollector.from_config(config,
                                             section['metrics_collector'])
        else:
            metrics_collector = None

        timeout = section.get('timeout')

        return cls(scoring_contexts, score_cache=score_cache,
                   metrics_collector=metrics_collector, timeout=timeout)


class TimeoutError(Exception):
    pass


def timeout(func, *args, seconds=None, **kwargs):
    if seconds is None:
        return func(*args, **kwargs)

    try:
        start = time.time()
        with stopit.ThreadingTimeout(seconds) as to_ctx_mgr:
            result = func(*args, **kwargs)
        duration = time.time() - start

        # OK, let's check what happened
        if to_ctx_mgr.state == to_ctx_mgr.EXECUTED:
            # All's fine, everything was executed
            return result
        elif to_ctx_mgr.state == to_ctx_mgr.TIMED_OUT:
            # Timeout occurred while executing the block
            raise TimeoutError("Timed out after {0} seconds.".format(duration))
        else:
            traceback.print_tb()
            raise RuntimeError("Something weird happened.")

    except stopit.TimeoutException as e:
        raise TimeoutError("Timed out after {0} seconds.".format(e.seconds))

    duration = time.time() - seconds



import logging
import socket

from .metrics_collector import MetricsCollector

logger = logging.getLogger(__name__)


class Statsd(MetricsCollector):
    """
    A metrics collector that uses :mod:`statsd` to connect to a graphite
    instance and send incr() and timing() events.
    """

    def __init__(self, statsd_client):
        self.statsd_client = statsd_client

    def precache_request(self, context, model, version, duration):
        with self.statsd_client.pipeline() as pipe:
            pipe.timing("precache_request.{0}.{1}.{2}"
                        .format(context, model, version),
                        duration * 1000)
            pipe.timing("precache_request.{0}.{1}".format(context, model),
                        duration * 1000)
            pipe.timing("precache_request.{0}".format(context),
                        duration * 1000)
            pipe.timing("precache_request",
                        duration * 1000)

    def scores_request(self, context, model, version, rev_id_count, duration):
        with self.statsd_client.pipeline() as pipe:
            pipe.timing("scores_request.{0}.{1}.{2}.{3}"
                        .format(context, model, version, rev_id_count),
                        duration * 1000)
            pipe.timing("scores_request.{0}.{1}.{2}"
                        .format(context, model, version),
                        duration * 1000)
            pipe.timing("scores_request.{0}.{1}".format(context, model),
                        duration * 1000)
            pipe.timing("scores_request.{0}".format(context),
                        duration * 1000)
            pipe.timing("scores_request",
                        duration * 1000)
            pipe.incr("revision_scored.{0}.{1}.{2}"
                        .format(context, model, version),
                      count=rev_id_count)
            pipe.incr("revision_scored.{0}.{1}".format(context, model),
                      count=rev_id_count)
            pipe.incr("revision_scored.{0}".format(context),
                      count=rev_id_count)
            pipe.incr("revision_scored", count=rev_id_count)

    def datasources_extracted(self, context, model, version, rev_id_count,
                              duration):
        with self.statsd_client.pipeline() as pipe:
            pipe.timing("datasources_extracted.{0}.{1}.{2}.{3}"
                        .format(context, model, version, rev_id_count),
                        duration * 1000)
            pipe.timing("datasources_extracted.{0}.{1}.{2}"
                        .format(context, model, version),
                        duration * 1000)
            pipe.timing("datasources_extracted.{0}.{1}"
                        .format(context, model),
                        duration * 1000)
            pipe.timing("datasources_extracted.{0}".format(context),
                        duration * 1000)
            pipe.timing("datasources_extracted",
                        duration * 1000)

    def score_processed(self, context, model, version, duration):
        with self.statsd_client.pipeline() as pipe:
            pipe.timing("score_processed.{0}.{1}.{2}"
                        .format(context, model, version),
                        duration * 1000)
            pipe.timing("score_processed.{0}.{1}".format(context, model),
                        duration * 1000)
            pipe.timing("score_processed.{0}".format(context),
                        duration * 1000)
            pipe.timing("score_processed", duration * 1000)

    def score_processor_overloaded(self, context, model, version, count=1):
        with self.statsd_client.pipeline() as pipe:
            pipe.incr("score_processor_overloaded.{0}.{1}.{2}"
                      .format(context, model, version),
                      count=count)
            pipe.incr("score_processor_overloaded.{0}.{1}"
                      .format(context, model),
                      count=count)
            pipe.incr("score_processor_overloaded.{0}".format(context),
                      count=count)
            pipe.incr("score_processor_overloaded",
                      count=count)

    def score_cache_hit(self, context, model, version, count=1):
        with self.statsd_client.pipeline() as pipe:
            pipe.incr("score_cache_hit.{0}.{1}.{2}"
                      .format(context, model, version),
                      count=count)
            pipe.incr("score_cache_hit.{0}.{1}".format(context, model),
                      count=count)
            pipe.incr("score_cache_hit.{0}".format(context), count=count)
            pipe.incr("score_cache_hit".format(context), count=count)

    def score_errored(self, context, model, version, count=1):
        with self.statsd_client.pipeline() as pipe:
            pipe.incr("score_errored.{0}.{1}.{2}"
                      .format(context, model, version),
                      count=count)
            pipe.incr("score_errored.{0}.{1}".format(context, model),
                      count=count)
            pipe.incr("score_errored.{0}".format(context), count=count)
            pipe.incr("score_errored".format(context), count=count)

    @classmethod
    def from_parameters(cls, *args, **kwargs):
        import statsd
        statsd_client = statsd.StatsClient(*args, **kwargs)
        return cls(statsd_client)

    @classmethod
    def from_config(cls, config, name, section_key="metrics_collectors"):
        """
        metrics_collectors:
            wmflabs_statsd:
                class: ores.metrics_collectors.Statsd
                host: labmon1001.eqiad.wmnet
                prefix: ores.{hostname}
                maxudpsize: 512
        """
        logger.info("Loading Statsd '{0}' from config.".format(name))
        section = config[section_key][name]

        kwargs = {k: v for k, v in section.items() if k != "class"}
        if 'prefix' in kwargs:
            kwargs['prefix'] = kwargs['prefix'] \
                               .format(hostname=socket.gethostname())
        return cls.from_parameters(**kwargs)



import json
import logging
from hashlib import sha1

from .score_cache import ScoreCache

logger = logging.getLogger("ores.score_caches.redis")

TTL = 60 * 60 * 24 * 365 * 16  # 16 years
PREFIX = "ores"


class Redis(ScoreCache):

    def __init__(self, redis, ttl=None, prefix=None):
        self.redis = redis
        self.ttl = int(ttl or TTL)
        self.prefix = str(prefix or PREFIX)

    def lookup(self, wiki, model, rev_id, version=None, cache=None):
        key = self._generate_key(wiki, model, rev_id, version=version,
                                 cache=cache)

        logger.debug("Looking up score at {0}".format(key))
        value = self.redis.get(key)
        if value is None:
            raise KeyError(key)
        else:
            return json.loads(str(value, 'utf-8'))

    def store(self, wiki, model, rev_id, score, version=None, cache=None):
        key = self._generate_key(wiki, model, rev_id, version=version,
                                 cache=cache)

        logger.debug("Storing score at {0}".format(key))
        self.redis.setex(key, self.ttl, bytes(json.dumps(score), 'utf-8'))

    def _generate_key(self, wiki, model, rev_id, version=None, cache=None):
        if cache is None or len(cache) == 0:
            key_values = [self.prefix, wiki, model, rev_id, version]
        else:
            cache_hash = self.hash_cache(cache)
            key_values = [self.prefix, wiki, model, rev_id, version,
                          cache_hash]

        return ":".join(str(v) for v in key_values)

    @classmethod
    def from_parameters(cls, *args, ttl=None, prefix=None, **kwargs):
        try:
            import redis
        except ImportError:
            raise ImportError("Could not find redis-py.  This packages is " +
                              "required when using ores.score_caches.Redis.")

        return cls(redis.StrictRedis(*args, **kwargs), ttl=ttl, prefix=prefix)

    @classmethod
    def from_config(cls, config, name, section_key="score_caches"):
        """
        score_caches:
            redis_cache:
                class: ores.score_caches.Redis
                host: localhost
                prefix: ores-derp
                ttl: 9001
        """
        logger.info("Loading Redis '{0}' from config.".format(name))
        section = config[section_key][name]

        kwargs = {k: v for k, v in section.items() if k != "class"}

        return cls.from_parameters(**kwargs)

    @classmethod
    def hash_cache(cls, cache):
        sorted_tuple = tuple(sorted(cache.items()))
        return sha1(bytes(str(sorted_tuple), 'utf8')).hexdigest()



"""
Starts a celery worker for ORES.  Note that

Usage:
    celery_worker -h | --help
    celery_worker [--config=<path>] [--verbose]

Options:
    -h --help          Prints this documentation
    --config=<path>    The path to a yaml config file
                       [default: config]
    --verbose          Turns up logging level
"""
import docopt
import glob
import logging
import os
import yamlconf

from ..score_processors import Celery


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    ores_level = logging.DEBUG if args['--verbose'] else logging.INFO
    logging.getLogger('ores').setLevel(ores_level)

    config_paths = os.path.join(args['--config'], "*.yaml")
    config = yamlconf.load(*(open(p) for p in
                             sorted(glob.glob(config_paths))))

    name = config['ores']['score_processor']
    score_processor = Celery.from_config(config, name)

    score_processor.application.worker_main(
        argv=["celery_worker", "--loglevel=INFO"]
    )



"""
Runs a uwsgi server.

Usage:
    wsgi [--config-dir=<path>]... [--logging-config=<path>]

Options:
    -h --help                Prints this documentation
    --config-dir=<path>      The path to a directory containing configuration
                             [default: config/]
    --logging-config=<path>  The path to a logging configuration file
"""
import logging

import docopt
from ores.wsgi import server

from .util import build_config


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    run(config_dirs=args['--config-dir'],
        logging_config=args['--logging-config'])


def run(*args, **kwargs):
    application = build(*args, **kwargs)
    logging.getLogger('ores').setLevel(logging.DEBUG)
    application.debug = True
    application.run(host="0.0.0.0", processes=64, debug=True)


def build(*args, **kwargs):
    config = build_config(*args, **kwargs)
    return server.configure(config)



"""
Runs a celery worker.

Usage:
    celery [--config-dir=<path>]... [--logging-config=<path>]

Options:
    -h --help                Prints this documentation
    --config-dir=<path>      The path to a directory containing configuration
                             [default: config/]
    --logging-config=<path>  The path to a logging configuration file
"""

import logging

import docopt

from ..score_processors import Celery
from .util import build_config


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    run(config_dirs=args['--config-dir'],
        logging_config=args['--logging-config'])


def run(*args, **kwargs):
    application = build(*args, **kwargs)
    logging.getLogger('ores').setLevel(logging.DEBUG)
    application.worker_main(argv=["celery_worker", "--loglevel=INFO"])


def build(*args, **kwargs):
    config = build_config(*args, **kwargs)
    score_processor = Celery.from_config(
        config, config['ores']['score_processor'])
    return score_processor.application



import logging

from .metrics_collector import MetricsCollector


class Logger(MetricsCollector):
    """
    A metrics collector that uses :mod:`logging` to report usage metrics.
    """
    def __init__(self, logger=None):
        self.logger = logger or logging.getLogger(__name__)

    def precache_request(self, context, model, version, duration):
        self.logger.debug("precache_request: {0}:{1}:{2} in {3} seconds"
                          .format(context, model, version, duration))

    def scores_request(self, context, model, version, rev_id_count, duration):
        self.logger.debug("scores_request: " +
                          "{0}:{1}:{2} for {3} revisions in {4} seconds"
                          .format(context, model, version, rev_id_count,
                                  duration))

    def datasources_extracted(self, context, model, version, rev_id_count, duration):
        self.logger.debug("datasources_extracted: " +
                          "{0}:{1}:{2} for {3} revisions in {4} seconds"
                          .format(context, model, version, rev_id_count,
                                  duration))

    def score_processor_overloaded(self, context, model, version, count=1):
        self.logger.debug("score_processor_overloaded: " +
                          "{0}:{1}:{2}".format(context, model, version))

    def score_processed(self, context, model, version, duration):
        self.logger.debug("score_processed: {0}:{1}:{2} in {3} seconds"
                          .format(context, model, version, duration))

    def score_cache_hit(self, context, model, version, count=1):
        for i in range(count):
            self.logger.debug("score_cache_hit: {0}:{1}:{2}"
                              .format(context, model, version))

    def score_errored(self, context, model, version, count=1):
        for i in range(count):
            self.logger.debug("score_errored: {0}:{1}:{2}"
                              .format(context, model, version))

    @classmethod
    def from_config(cls, config, name, section_key="metrics_collectors"):
        """
        metrics_collectors:
            local_logging:
                class: ores.metrics_collectors.Logging
        """
        return cls()



import os

import yaml
from flask.ext.jsonpify import jsonify


def configure(config, bp, score_processor):

    # /spec/
    @bp.route("/v2/spec/", methods=["GET"])
    def v2_spec():
        return generate_spec()

    return bp


def generate_spec():
    dir_name = os.path.dirname(os.path.abspath(__file__))
    swagger_doc = yaml.load(open(os.path.join(dir_name, "swagger.yaml")))
    return jsonify(swagger_doc)



from flask import request
from flask.ext.jsonpify import jsonify

from ... import responses
from .... import errors
from ...util import (CacheParsingError, ParamError, parse_injection,
                     read_bar_split_param)


def configure(config, bp, score_processor):

    # /v2/scores/
    @bp.route("/v2/scores/", methods=["GET"])
    def scores_v2():
        scores_doc = {}

        # model_info param
        if 'model_info' in request.args:
            try:
                model_info_fields = set(read_bar_split_param(request, "model_info", type=str))
            except ParamError as e:
                return responses.bad_request(str(e))
        else:
            model_info_fields = None

        for context, scoring_context in score_processor.items():
            context_doc = {}
            for model, scorer_model in scoring_context.items():
                model_doc = {'version': scorer_model.version}

                if model_info_fields is not None:
                    info_doc = scorer_model.format_info(format='json')
                    if len(model_info_fields) > 0 and \
                       model_info_fields != {''}:
                        info_doc = {k: info_doc[k]
                                    for k in model_info_fields
                                    if k in info_doc}

                    model_doc['info'] = info_doc

                context_doc[model] = model_doc

            scores_doc[context] = context_doc

        return jsonify({'scores': scores_doc})

    # /v2/scores/enwiki/?models=reverted&revids=456789|4567890
    @bp.route("/v2/scores/<context>/", methods=["GET"])
    def score_model_revisions_v2(context):
        scores_doc = {context: {}}
        # Check to see if we have the context available in our score_processor
        if context not in score_processor:
            return responses.not_found("No scorers available for {0}"
                                       .format(context))

        # models param
        if "models" in request.args:
            # Read the models param
            try:
                models = set(read_bar_split_param(request, "models", type=str))
            except ParamError as e:
                return responses.bad_request(str(e))

            # Check if all the models are available
            missing_models = models - score_processor[context].keys()
            if len(missing_models) > 0:
                return responses.bad_request("Models '{0}' not available for {1}."
                                             .format(list(missing_models), context))
        else:
            # Empty?  All the models!
            models = score_processor[context].keys()

        for model in models:
            scores_doc[context][model] = \
                {'version': score_processor[context][model].version}

        # model_info param
        if 'model_info' in request.args:
            try:
                model_info_fields = set(read_bar_split_param(request, "model_info", type=str))
            except ParamError as e:
                return responses.bad_request(str(e))

            for model in models:
                model_info_doc = \
                    score_processor[context][model].format_info(format='json')
                if len(model_info_fields) > 0 and model_info_fields != {''}:
                    model_info_doc = {k: model_info_doc[k]
                                      for k in model_info_fields
                                      if k in model_info_doc}

                scores_doc[context][model]['info'] = model_info_doc

        # Read the rev_ids
        try:
            rev_ids = set(read_bar_split_param(request, "revids", type=int))
        except ParamError as e:
            return responses.bad_request(str(e))

        if len(rev_ids) > 0:
            precache = "precache" in request.args

            try:
                for model in models:
                    model_scores, _ = score_processor.score(
                        context, model, rev_ids, precache=precache)
                    scores_doc[context][model]['scores'] = model_scores
            except errors.ScoreProcessorOverloaded:
                return responses.server_overloaded()

        return jsonify({'scores': scores_doc})

    # /v2/scores/enwiki/reverted/?revids=456789|4567890
    @bp.route("/v2/scores/<context>/<model>/", methods=["GET"])
    def score_revisions_v2(context, model):
        scores_doc = {context: {model: {}}}
        # Check to see if we have the context available in our score_processor
        if context not in score_processor:
            return responses.not_found("No models available for {0}"
                                       .format(context))

        if model not in score_processor[context]:
            return responses.bad_request("Model '{0}' not available for {1}."
                                         .format(model, context))

        scores_doc[context][model] = \
            {'version': score_processor[context][model].version}

        if 'model_info' in request.args:
            try:
                model_info_fields = set(
                    read_bar_split_param(request, "model_info", type=str))
            except ParamError as e:
                return responses.bad_request(str(e))

            model_info_doc = \
                score_processor[context][model].format_info(format='json')
            if len(model_info_fields) > 0 and model_info_fields != {''}:
                model_info_doc = {k: model_info_doc[k]
                                  for k in model_info_fields
                                  if k in model_info_doc}

            scores_doc[context][model]['info'] = model_info_doc

        # Read the rev_ids
        try:
            rev_ids = set(read_bar_split_param(request, "revids", type=int))
        except ParamError as e:
            return responses.bad_request(str(e))

        if len(rev_ids) > 0:
            precache = "precache" in request.args

            try:
                model_scores, _ = score_processor.score(
                    context, model, rev_ids, precache=precache)
                scores_doc[context][model]['scores'] = model_scores
            except errors.ScoreProcessorOverloaded:
                return responses.server_overloaded()

        return jsonify({'scores': scores_doc})

    # /v2/scores/enwiki/reverted/4567890
    @bp.route("/v2/scores/<context>/<model>/<int:rev_id>/", methods=["GET", "POST"])
    def score_revision_v2(context, model, rev_id):
        scores_doc = {context: {model: {}}}
        # Check to see if we have the context available in our score_processor
        if context not in score_processor:
            return responses.not_found("No models available for {0}"
                                       .format(context))

        if model not in score_processor[context]:
            return responses.bad_request("Model '{0}' not available for {1}."
                                         .format(model, context))

        scores_doc[context][model] = \
            {'version': score_processor[context][model].version}

        if 'model_info' in request.args:
            try:
                model_info_fields = set(
                    read_bar_split_param(request, "model_info", type=str))
            except ParamError as e:
                return responses.bad_request(str(e))

            model_info_doc = \
                score_processor[context][model].format_info(format='json')
            if len(model_info_fields) > 0 and model_info_fields != {''}:
                model_info_doc = {k: model_info_doc[k]
                                  for k in model_info_fields
                                  if k in model_info_doc}

            scores_doc[context][model]['info'] = model_info_doc

        try:
            caches = parse_injection(request, rev_id)
        except CacheParsingError as e:
            return responses.bad_request("Unabled to parse params: {0}"
                                         .format(e))

        include_features = "features" in request.args
        precache = "precache" in request.args

        try:
            model_scores, feature_values = score_processor.score(
                context, model, [rev_id], caches=caches,
                precache=precache, include_features=include_features)
            scores_doc[context][model]['scores'] = model_scores
            if include_features:
                scores_doc[context][model]['features'] = feature_values
        except errors.ScoreProcessorOverloaded:
            return responses.server_overloaded()

        return jsonify({'scores': scores_doc})

    return bp



from ..null import Null


def test_null():
    # Make sure we throw no errors.

    collector = Null()
    collector.precache_request("foo", "bar", "0.0.1", 100)
    collector.scores_request("foo", "bar", "0.0.1", 50, 150)
    collector.datasources_extracted("foo", "bar", "0.0.1", 10, 25)
    collector.score_processed("foo", "bar", "0.0.1", 1.1)
    collector.score_cache_hit("foo", "bar", "0.0.1", 2)
    collector.score_errored("foo", "bar", "0.0.1")


def test_from_config():
    # Should throw a socket connection error and no others
    config = {
        'metrics_collectors': {
            'null': {
                'class': 'ores.metrics_collectors.Null'
            }
        }
    }
    Null.from_config(config, 'null')



import logging
import time

from revscoring import dependencies
from revscoring.datasources import Datasource
from revscoring.extractors import Extractor
from revscoring.features import trim
from revscoring.scorer_models import ScorerModel

logger = logging.getLogger(__name__)


class ScoringContext(dict):

    def __init__(self, name, scorer_models, extractor):
        """
        :Parameters:
            name : str
               The  that this scorer is being used for
            scorer_models : dict
               A mapping between names and
               :class:`revscoring.ScorerModel`
               instances
            extractor : :class:`revscoring.Extractor`
               An extractor to use for gathering feature values
        """
        super().__init__()
        self.name = str(name)
        self.update(scorer_models)
        self.extractor = extractor

    def solve_features(self, model, cache=None):
        """
        Solves a model's features for a revision.  Does not attempt to gather
        new data.
        """
        features = self[model].features
        return self.extractor.solve(features, cache=cache)

    def solve_base_features(self, model, cache=None):
        """
        Solves a model's basic features for a revision.  Does not attempt to
        gather new data.
        """
        features = list(trim(self[model].features))
        feature_values = self.extractor.solve(features, cache=cache)
        return {"feature." + f.name: v
                for f, v in zip(features, feature_values)}

    def version(self, model):
        return self[model].version

    def score(self, model, cache=None, include_features=False):
        """
        I am the score function
        """
        version = self[model].version

        # TODO: record time spend computing features
        start = time.time()
        feature_values = list(self.solve_features(model, cache))
        logger.debug("Extracted features for {0}:{1}:{2} in {3} secs"
                     .format(self.name, model, version, time.time() - start))

        # TODO: record time spent generating a score
        start = time.time()
        score = self[model].score(feature_values)
        logger.debug("Scored features for {0}:{1}:{2} in {3} secs"
                     .format(self.name, model, version, time.time() - start))

        if include_features:
            start = time.time()
            feature_vals = self.solve_base_features(model, cache)
            logger.debug("Re-extracted base features for " +
                         "{0}:{1}:{2} in {3} secs"
                         .format(self.name, model, version, time.time() - start))
        else:
            feature_vals = None

        return score, feature_vals

    def extract_roots(self, model, rev_ids, caches=None):
        """
        Extracts a mapping of root datasources capabile of generating the
        features needed for a particular model.

        :Parameters:
            rev_ids : int | `iterable`
                Revision IDs to extract for
            model : str
                The name of a :class:`~revscoring.ScorerModel` to
                extract the roots for
        """
        features = self[model].features
        root_ds = [d for d in dependencies.dig(features)
                   if isinstance(d, Datasource)]
        error_root_vals = self.extractor.extract(rev_ids, root_ds,
                                                 caches=caches)
        root_ds_caches = {}
        for rev_id, (error, root_vals) in zip(rev_ids, error_root_vals):
            if error is None:
                root_ds_map = {rd: rv for rd, rv in zip(root_ds, root_vals)}
                if caches is not None:
                    root_ds_map.update(caches.get(rev_id, {}))
                root_ds_caches[rev_id] = (None, root_ds_map)
            else:
                root_ds_caches[rev_id] = (error, None)

        return root_ds_caches

    @classmethod
    def from_config(cls, config, name, section_key="scoring_contexts"):
        """
        Expects:

            scoring_contexts:
                enwiki:
                    scorer_models:
                        damaging: enwiki_damaging_2014
                        good-faith: enwiki_good-faith_2014
                    extractor: enwiki
                ptwiki:
                    scorer_models:
                        damaging: ptwiki_damaging_2014
                        good-faith: ptwiki_good-faith_2014
                    extractor: ptwiki

            extractors:
                enwiki_api: ...
                ptwiki_api: ...

            scorer_models:
                enwiki_damaging_2014: ...
                enwiki_good-faith_2014: ...
        """
        logger.info("Loading ScoringContext '{0}' from config.".format(name))
        section = config[section_key][name]

        scorer_models = {}
        for model_name, key in section['scorer_models'].items():
            scorer_model = ScorerModel.from_config(config, key)
            scorer_models[model_name] = scorer_model

        extractor = Extractor.from_config(config, section['extractor'])

        return cls(name, scorer_models=scorer_models, extractor=extractor)



"""
From https://gist.github.com/Spindel/1d07533ef94a4589d348

watchdogged.py; python teaching code for how to use the systemd watchdog
Copyright (C) 2015 D.S. Ljungmark, Modio AB

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import socket


def notify_socket(clean_environment=True):
    """Return a tuple of address, socket for future use.

    clean_environment removes the variables from env to prevent children
    from inheriting it and doing something wrong.
    """
    _empty = None, None
    address = os.environ.get("NOTIFY_SOCKET", None)
    if clean_environment:
        address = os.environ.pop("NOTIFY_SOCKET", None)

    if not address:
        return _empty

    if len(address) == 1:
        return _empty

    if address[0] not in ("@", "/"):
        return _empty

    if address[0] == "@":
        address = "\0" + address[1:]

    # SOCK_CLOEXEC was added in Python 3.2 and requires Linux >= 2.6.27.
    # It means "close this socket after fork/exec()
    try:
        sock = socket.socket(socket.AF_UNIX,
                             socket.SOCK_DGRAM | socket.SOCK_CLOEXEC)
    except AttributeError:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

    return address, sock


def sd_message(address, sock, message):
    """Send a message to the systemd bus/socket.

    message is expected to be bytes.
    """
    if not (address and sock and message):
        return False
    assert isinstance(message, bytes)

    try:
        retval = sock.sendto(message, address)
    except socket.error:
        return False
    return (retval > 0)


def watchdog_ping(address, sock):
    """Helper function to send a watchdog ping."""
    message = b"WATCHDOG=1"
    return sd_message(address, sock, message)



import os
import re

from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def requirements(fname):
    """
    Generator to parse requirements.txt file

    Supports bits of extended pip format (git urls)
    """
    with open(fname) as f:
        for line in f:
            match = re.search('#egg=(.*)$', line)
            if match:
                yield match.groups()[0]
            else:
                yield line.strip()


setup(
    name="ores",
    version="0.7.4",  # Update in ores/__init__.py too.
    author="Aaron Halfaker",
    author_email="ahalfaker@wikimedia.org",
    description=("A webserver for hosting scorer models."),
    license="MIT",
    entry_points={
        'console_scripts': [
            'ores = ores.ores:main',
        ],
    },
    url="https://github.com/wiki-ai/ores",
    packages=find_packages(),
    include_package_data=True,
    long_description=read('README.md'),
    install_requires=list(requirements("requirements.txt")),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)



from collections import defaultdict

from flask import request
from flask.ext.jsonpify import jsonify

from ... import responses
from .... import errors
from ...util import ParamError, read_bar_split_param


def configure(config, bp, score_processor):

    # /scores/
    @bp.route("/scores/", methods=["GET"])
    @bp.route("/v1/scores/", methods=["GET"])
    def scores():
        contexts = [context for context in score_processor]

        contexts.sort()

        return jsonify({'contexts': contexts})

    # /scores/enwiki/?models=reverted&revids=456789|4567890
    @bp.route("/scores/<context>/", methods=["GET"])
    @bp.route("/v1/scores/<context>/", methods=["GET"])
    def score_model_revisions(context):

        # Check to see if we have the context available in our score_processor
        if context not in score_processor:
            return responses.not_found("No scorers available for {0}"
                                       .format(context))

        # If no model is specified, return information on available models
        if "models" not in request.args:
            # Return the models that we have
            models = {name: model.format_info(format="json")
                      for name, model in score_processor[context].items()}
            return jsonify({"models": models})

        # Read the params
        try:
            models = set(read_bar_split_param(request, "models", type=str))
        except ParamError as e:
            return responses.bad_request(str(e))

        # Check if all the models are available
        missing_models = models - score_processor[context].keys()
        if len(missing_models) > 0:
            return responses.bad_request("Models '{0}' not available for {1}."
                                         .format(list(missing_models), context))

        # Read the rev_ids
        try:
            rev_ids = set(read_bar_split_param(request, "revids", type=int))
        except ParamError as e:
            return responses.bad_request(str(e))

        if len(rev_ids) == 0:
            return responses.bad_request("No revids provided.")

        precache = "precache" in request.args

        # Generate scores for each model and merge them together
        try:
            scores = defaultdict(dict)
            for model in models:
                model_scores, _ = score_processor.score(
                    context, model, rev_ids, precache=precache)
                for rev_id in model_scores:
                    scores[rev_id][model] = model_scores[rev_id]
        except errors.ScoreProcessorOverloaded:
            return responses.server_overloaded()

        return jsonify(scores)

    # /scores/enwiki/reverted/?revids=456789|4567890
    @bp.route("/scores/<context>/<model>/", methods=["GET"])
    @bp.route("/v1/scores/<context>/<model>/", methods=["GET"])
    def score_revisions(context, model):

        # Check to see if we have the context available in our score_processor
        if context not in score_processor:
            return responses.not_found("No models available for {0}"
                                       .format(context))

        if model not in score_processor[context]:
            return responses.bad_request("Model '{0}' not available for {1}."
                                         .format(model, context))

        # Try to read the rev_ids
        if 'revids' in request.args:
            try:
                rev_ids = set(read_bar_split_param(request, "revids", type=int))
            except ParamError as e:
                return responses.bad_request(str(e))

            if len(rev_ids) == 0:
                return responses.bad_request("No revids provided.")
        else:
            return jsonify(score_processor[context][model].format_info(format="json"))

        precache = "precache" in request.args
        try:
            model_scores, _ = score_processor.score(context, model, rev_ids,
                                                    precache=precache)
        except errors.ScoreProcessorOverloaded:
            return responses.server_overloaded()
        return jsonify(model_scores)

    # /scores/enwiki/reverted/4567890
    @bp.route("/scores/<context>/<model>/<int:rev_id>/", methods=["GET", "POST"])
    @bp.route("/v1/scores/<context>/<model>/<int:rev_id>/", methods=["GET", "POST"])
    def score_revision(context, model, rev_id):

        # Check to see if we have the context available in our score_processor
        if context not in score_processor:
            return responses.not_found("No models available for {0}"
                                       .format(context))

        if model not in score_processor[context]:
            return responses.not_found("Model '{0}' not available for {1}."
                                       .format(model, context))

        precache = "precache" in request.args

        try:
            model_scores, _ = score_processor.score(context, model, [rev_id],
                                                    precache=precache)
        except errors.ScoreProcessorOverloaded:
            return responses.server_overloaded()

        return jsonify(model_scores)

    return bp



import socket
from contextlib import contextmanager

from nose.tools import eq_, raises

from ..statsd import Statsd


def test_statsd():
    class StatsClient:

        def __init__(self):
            self.messages = []

        def incr(self, name, count=1):
            self.messages.append(("INCR", name, count))

        def timing(self, name, duration):
            self.messages.append(("TIMING", name, duration))

        @contextmanager
        def pipeline(self):
            yield self

    fake_client = StatsClient()

    collector = Statsd(fake_client)
    collector.precache_request("foo", "bar", "0.0.1", 100)
    collector.scores_request("foo", "bar", "0.0.1", 50, 150)
    collector.datasources_extracted("foo", "bar", "0.0.1", 10, 25)
    collector.score_processed("foo", "bar", "0.0.1", 1.1)
    collector.score_cache_hit("foo", "bar", "0.0.1", 2)
    collector.score_errored("foo", "bar", "0.0.1")

    eq_(fake_client.messages,
        [('TIMING', 'precache_request.foo.bar.0.0.1', 100000),
         ('TIMING', 'precache_request.foo.bar', 100000),
         ('TIMING', 'precache_request.foo', 100000),
         ('TIMING', 'precache_request', 100000),
         ('TIMING', 'scores_request.foo.bar.0.0.1.50', 150000),
         ('TIMING', 'scores_request.foo.bar.0.0.1', 150000),
         ('TIMING', 'scores_request.foo.bar', 150000),
         ('TIMING', 'scores_request.foo', 150000),
         ('TIMING', 'scores_request', 150000),
         ('INCR', 'revision_scored.foo.bar.0.0.1', 50),
         ('INCR', 'revision_scored.foo.bar', 50),
         ('INCR', 'revision_scored.foo', 50),
         ('INCR', 'revision_scored', 50),
         ('TIMING', 'datasources_extracted.foo.bar.0.0.1.10', 25000),
         ('TIMING', 'datasources_extracted.foo.bar.0.0.1', 25000),
         ('TIMING', 'datasources_extracted.foo.bar', 25000),
         ('TIMING', 'datasources_extracted.foo', 25000),
         ('TIMING', 'datasources_extracted', 25000),
         ('TIMING', 'score_processed.foo.bar.0.0.1', 1100),
         ('TIMING', 'score_processed.foo.bar', 1100),
         ('TIMING', 'score_processed.foo', 1100),
         ('TIMING', 'score_processed', 1100),
         ('INCR', 'score_cache_hit.foo.bar.0.0.1', 2),
         ('INCR', 'score_cache_hit.foo.bar', 2),
         ('INCR', 'score_cache_hit.foo', 2),
         ('INCR', 'score_cache_hit', 2),
         ('INCR', 'score_errored.foo.bar.0.0.1', 1),
         ('INCR', 'score_errored.foo.bar', 1),
         ('INCR', 'score_errored.foo', 1),
         ('INCR', 'score_errored', 1)])


@raises(socket.gaierror)
def test_from_config():
    # Should throw a socket connection error and no others
    config = {
        'metrics_collectors': {
            'wmflabs_statsd': {
                'class': 'ores.metrics_collectors.Statsd',
                'host': 'totally.doesnt.exists.OMG',
                'prefix': 'ores.{hostname}',
                'maxudpsize': 512
            }
        }
    }
    Statsd.from_config(config, 'wmflabs_statsd')



"""
Starts a development web server.

Usage:
    dev_server (-h | --help)
    dev_server [--host=<name>] [--port=<num>] [--config=<path>]
               [--processes=<num>] [--debug] [--verbose]

Options:
    -h --help          Print this documentation
    --host=<name>      The hostname to listen on [default: 0.0.0.0]
    --port=<num>       The port number to start the server on [default: 8080]
    --config=<path>    The path to a yaml config file
                       [default: config]
    --processes=<num>  The number of parallel processes to handle [default: 32]
    --debug            Print debug logging information
    --verbose          Print verbose extraction information
"""
import docopt
import glob
import logging
import os
import yamlconf

from ..wsgi import server


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    config_paths = os.path.join(args['--config'], "*.yaml")
    config = yamlconf.load(*(open(p) for p in
                             sorted(glob.glob(config_paths))))

    processes = int(args['--processes'])

    verbose = args['--verbose']

    debug = args['--debug']

    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    logging.getLogger('requests').setLevel(logging.INFO)
    if verbose:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.DEBUG)
    else:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.INFO)

    logging.getLogger("ores.metrics_collectors.logger").setLevel(logging.DEBUG)

    app = server.configure(config)
    app.run(host=args['--host'],
            port=int(args['--port']),
            debug=True,
            processes=processes)



import time

from nose.tools import eq_
from revscoring.features import Feature
from revscoring.scorer_models import ScorerModel

from ...scoring_contexts import ScoringContext
from ..timeout import Timeout

wait_time = Feature("wait_time", returns=float)


class FakeSM(ScorerModel):

    def __init__(self):
        self.features = [wait_time]
        self.language = None
        self.version = None

    def score(self, feature_values):
        raise NotImplementedError()


class FakeSC(ScoringContext):

    def solve(self, model, cache, include_features=False):
        return cache

    def score(self, model, cache, include_features=False):
        wait_time_value = cache[wait_time]
        time.sleep(wait_time_value)
        return {'score': True}, None

    def extract_roots(self, model, rev_ids, caches=None):
        return {rev_id: (None, caches[rev_id]) for rev_id in rev_ids}


def test_score():
    fakewiki = FakeSC("fakewiki", {'fake': FakeSM()}, None)
    score_processor = Timeout({'fakewiki': fakewiki}, timeout=0.10)

    scores, feature_maps = score_processor.score(
            "fakewiki", "fake", [1], caches={1: {wait_time: 0.05}})
    eq_(scores, {1: {'score': True}})


def test_timeout():
    fakewiki = FakeSC("fakewiki", {'fake': FakeSM()}, None)
    score_processor = Timeout({'fakewiki': fakewiki}, timeout=0.05)

    scores, feature_maps = score_processor.score(
        "fakewiki", "fake", [1], caches={1: {wait_time: 0.10}})
    assert 'error' in scores[1]
    assert 'Timed out after' in scores[1]['error']['message']
